#!/usr/bin/python
# A pseudo protocol for PyPseudoserver
# doesnt work on its own, you should extend this module with your own protocol

from twisted.internet import reactor, task
from twisted.words.protocols.irc import IRC
import MySQLdb as db
import sys, os, re, commands as pycmd
sys.path.insert(0, "networkx-0.99-py2.5.egg")
import networkx as nx
from __main__ import *

class PseudoProtocol(IRC):
	"""A PseudoServer based on the IRCD protocol
	This should be extended out for your own protocol if you write one.
	"""
	hostname = config.get('server', 'name')
	modules = {}
	hooks = {}
	commands = {}
	burst = True
	first_burst = True
	flood = False
	last_count = 0
	logchan = config.get('control', 'channel')
	# pycmd.getoutput("hg heads") or ? allows this to continue to load on machines where the hg executable  is not in
	# the local directory of the system search path
	try:
		version = "PyPseudoServer.r" + re.match("^changeset: +(\d+):.*", pycmd.getoutput("hg heads")).groups()[0]
	except:
		version = "PyPseudoServer.r ?"
	masters = config.get('control', 'master').split(",")
	dbp = None #persistent datastore
	
	"""
	The server_t class variable contains information about the servers
	in the network.  It's a dict of dicts, to allow for some flexibility
	when developing modules and protocols.
	KEYS: TS6 networks are keyed by SID, otherwise key by 'host'
	Standard subkeys are as follows:
	'host' = server hostname (irc.mycool.net)
	'desc' = server description (Cold as ICE, baby)
	'uplink' = server that connects to this server (hub.mycool.net)
	'hops' = how many hops to that server (2)
	"""
	server_t = {} # server table
	server_tree = None #server link tree (used with graphing and networkx)
	
	"""
	The user_t class variable contains information about users
	on the network and "my" private users from modules and control users, respectively.
	KEYS: TS6 networks are keyed by UID, otherwise key by 'Nick'
	Standard subkeys are as follows:
	'nick' = user nickname (g0d)
	'user' = user field of the mask (~mirc32)
	'host' = host field of the mask (MyCoolNET-D3FDS3A.21.221.myisp.net) [rdns or spoof allowed]
	'gecos' = gecos of the user (I am awesome)
	'hops' = hops to the user (4)
	'modes' = umodes set on the user (+iwx)
	'ip' = ip address of the user (55.23.21.221)
	'rdns' = reverse dns of the ip (dialup-1.55.23.21.221.myisp.net)
	'server' = server host (must match 'host' in server_t) that the user is on
	'version' = client version response, default is ""
	ADDITIONALLY, my_user_t must contain:
	'nspass' = string/password to identify to NickServ with
	"""
	user_t = {} # user table
	
	"""
	The channel_t class variable contains channel related information for the network,
	including full user->channels and channel->users maps.
	KEYS: keyed by name (#channel)
	Standard subkeys are as follows:
	'modes' = modes set on the channel (+stnl)
	'limit' = if +l is set, this is the channel limit
	'key' = if +k is set, this is the channel key
	'topic' = current topic of the channel
	'members' = list of users (nick or uid)
	"""
	channel_t = {} # channel table

	
	def __init__(self):
		"""Initialize the system"""
		
		try:
			self.dbx = db.connect(
				host=config.get('database', 'host'),
				user=config.get('database', 'user'),
				passwd=config.get('database', 'passwd'),
				db=config.get('database', 'db'),
				init_command='SET NAMES utf8',
				use_unicode=True,
			)
			self.dbx.autocommit(True) #no need to have transactions
			self.dbp = self.dbx.cursor()
			self.dbp.execute("CREATE TABLE IF NOT EXISTS record (name VARCHAR(255), num INT(10), UNIQUE KEY (name));")
		except Exception, err:
			raise Exception("Error initializing database (%s), exiting..." % err)
		
		try:
			self.flood_check_usermark = config.getint('control', 'flood_usermark')
		except Exception, err:
			raise Exception("Error getting config settings: %s" % err)
		try:
			self.server_tree = nx.Graph()
		except Exception, err:
			raise Exception("Error instantiating server tree graph (%s), exiting..." % err)
		
	def c2u(self, data):
		"""Attempts to convert text to unicode"""
		encodings = ["ascii", "iso-2022-jp", "utf-8", "euc-jp", "sjis", "iso-8859-1"]
		if type(data).__name__ == "unicode": return data
		for enc in encodings:
			try:
				text = unicode(data, enc)
				break
			except (UnicodeDecodeError, UnicodeError):
				enc = ""
				text = ""
		return text
		
	def reload(self):
		"""Re-read the configuration file and reload some values
		"""
		
		try:
			config.read("config.ini")
			self.masters = config.get('control', 'master').split(",")
			self.logger.loglevel = config.getint('logging', 'level')
			self.flood_check_usermark = config.getint('control', 'flood_usermark')
			return True
		except Exception, err:
			self.logger.log("Error reloading config.ini (%s)" % err, L_ERROR)
			return False
		
	def shutdown(self):
		"""Exit the system"""
		
		for name in self.modules.copy():
			try:
				self.logger.log("Shutting down module %s" % name, L_DEBUG)
				self.unloadModule(name)
			except Exception, err:
				self.logger.log("Error shutting down module %s (%s)" % (name, err), L_ERROR)
	
	def loadModule(self, name):
		"""Load the specified module into the runtime code.
		Also register any hooks and commands it has.
		Note, this function changed from accepting a list to accepting a single string.
		To duplicate old functionality, loop your call.
		"""
		
		if name in self.modules:
			self.logger.log("Error loading module %s, exists in global table already" % name, L_ERROR)
			return False
		
		self.logger.log("Loading: %s" % name, L_STATUS)
		
		try:
			mnproper = "psm_" + name
			mod = __import__("modules." + mnproper)
			mod = eval("mod." + mnproper)
			obj = getattr(mod, "PSModule_" + name)(self, config)
		except ImportError, err:
			self.logger.log("Error importing: %s (%s)" % (name, err), L_ERROR)
			return False
		except Exception, err:
			self.logger.log("Error instantiating(a) module %s (%s)" % (name, err), L_ERROR)
			return False
			
		self.modules[name] = obj
		
		try:
			for hook, cb in obj.getHooks():
				self.registerHook(hook, cb)
			for hook, args in obj.getCommands():
				self.registerCommand(hook, args)
			if not self.burst:
				obj.startup()
			return obj.getVersion() 
		except Exception, err:
			self.logger.log("Error instantiating(b) %s (%s)" % (name, err), L_ERROR)
			self.unloadModule(name)
			return False
	
	def unloadModule(self, name):
		"""Unload the module specified from runtime code
		and the global module list.
		TODO make this throw exceptions instead of return False
		"""
		
		if name not in self.modules:
			self.logger.log("Module %s failed to unload: not in global table" % name, L_ERROR)
			return False

		obj = self.modules[name]

		try:
			for hook,cb in obj.getHooks():
				try:
					self.logger.log("Unregistering hook %s for %s" % (hook,cb), L_DEBUG)
					self.hooks[hook].remove(cb)
				except Exception, err:
					self.logger.log("Error unloading hook %s for module %s (%s)" % (hook,name,err), L_ERROR)
		except AttributeError, err:
			self.logger.log("Error unloading hooks: %s" % err, L_ERROR)

		try:
			for hook,args in obj.getCommands():
				try:
					self.logger.log("Unregistering command %s" % (hook), L_DEBUG)
					del self.commands[hook]
				except KeyError, err:
					self.logger.log("Error unloading command %s for module %s (%s)" % (hook,name,err), L_ERROR)
		except AttributeError, err:
			self.logger.log("Error unloading commands: %s" % err, L_ERROR)

		try:
			obj.shutdown()
		except Exception, err:
			self.logger.log("Error shutting down module %s (%s)" % (name,err), L_ERROR)
			return False
	
		#sometimes the module is imported, but we need this here to unload
		# if something bad happens
		try:
			sys.modules.pop("modules.psm_%s" % name)
		except Exception, err:
			self.logger.log("Error popping module %s off the import stack" % name, L_ERROR)
		
		del self.modules[name]
		return True
			
	def startModules(self):
		"""Execute the startup() method of all loaded modules.
		Modules should check that they've been started already
		prior to complying.
		"""
		for name,obj in self.modules.copy().iteritems():
			try:
				if not obj.startup():
					self.logger.log("Module %s was already started" % name, L_ERROR)
			except Exception, err:
				self.logger.log("Error starting module %s (%s), unloading..." % (name,err), L_ERROR)
				self.unloadModule(name)
	
	def registerHook(self, hook_name, callback):
		"""Register a hook from a module
		"""
		
		self.logger.log("Registered hook %s for %s" % (hook_name, callback), L_DEBUG)
		if not self.hooks.has_key(hook_name):
			self.hooks.update({hook_name:[]})
			
		self.hooks[hook_name].append(callback)
	
	def registerCommand(self, command_name, command_dict):
		"""Register a PRIVMSG command
		For main and modules
		"""
		
		self.logger.log("Registered command %s" % (command_name), L_DEBUG)
		if self.commands.has_key(command_name):
			self.logger.log("Command '%s' has already been registered, overwriting." % command_name, L_STATUS)
		self.commands.update({command_name:command_dict})
	
	def callHooks(self, hook_name, prefix, params):
		"""Called in the main thread to execute module hooks
		"""
		
		if not hook_name in self.hooks: return
		for hooks in self.hooks[hook_name]:
			self.logger.log("Trying to call hook %s with %s %s" % (hooks, prefix, params), L_DEBUG)
			try:
				hooks(prefix, params)
			except Exception, err:
				self.logger.log("Error calling hook: %s|%s|%s|%s" % (hooks, prefix, params, err), L_ERROR)

	def connectionMade(self):
		"""Called when a connection is established to the uplink"""
		
		IRC.connectionMade(self)
		self.logger = self.factory.logger
		
		self.logger.log('Connected to uplink', L_STATUS)
		
		# Load modules
		modules = config.get('modules', 'autoload')
		if modules.isspace() or len(modules) == 0: mlist = []
		else: mlist = modules.split(",")
		try:
			map(self.loadModule, mlist)
		except Exception, err:
			self.logger.log("Error loading modules (%s)" % err, L_ERROR)

		self.flood_timer = task.LoopingCall(self.check_flood)
		self.flood_timer.start(1, now=False)
	
	def connectionLost(self, reason):
		"""Called when a connection is lost to the uplink"""
		
		self.logger.log('Disconnected from uplink', L_STATUS)
		self.logger.close()
		
#### override parent method
	def sendLine(self, line):
		"""Overrides standard line sender
		Modification allows optional logging of outgoing data
		Also converts to utf-8 to avoid line errors.
		"""
		
		text = line.encode('utf-8')
		self.logger.log('>>> %s' % text, L_OUT)
		IRC.sendLine(self, text)
		
	def handleCommand(self, command, prefix, params):
		"""Overrides standard command handler
		Modification allows optional logging of input data
		"""
		
		try:
			command = self.c2u(command)
			prefix = self.c2u(prefix)
			params = map(lambda x: self.c2u(x), params)
		except:
			self.logger.log("ERROR: handleCommand() unicode decoding", L_ERROR)
			raise
		
		self.logger.log('[%s](%s): %s' % (command, prefix, '|'.join(params)), L_IN)
		IRC.handleCommand(self, command, prefix, params)
	
	def irc_unknown(self, prefix, command, params):
		"""Overrides handler for methods that don't exist
		Default raises exceptions, this does not.
		"""
		
		#away falls here	
		# call a handler, maybe a module implements it
		self.callHooks(command, prefix, params)
		
		pass
### Begin IRC hooks
	def irc_ERROR(self, prefix, params):
		"""Handle ERROR messages from uplink
		ERROR signifies something bad happened, so just log and drop the link
		"""
		
		reactor.stop()

	def irc_PING(self, prefix, params):
		"""Handle pings, signal end of burst by receiving a ping"""
		
		self.sendMessage("PONG", self.hostname, params[0], prefix=self.hostname)
		if self.burst:
			self.burst = False

	def irc_VERSION(self, prefix, params):
		"""Respond to VERSION requests sanely"""
		
		self.sendMessage("351", "%s %s %s :feat=%s" %
			(prefix, self.version, self.hostname, "+".join(self.modules)))
### End IRC hooks
	def check_flood(self):
		"""Compare user count changes and signal if flood is detected"""
		
		# don't check during a burst, duh.
		if self.burst:
			self.last_count = 0
			return
		
		current_count = len(self.user_t)
		
		if self.last_count:
			if (current_count-self.last_count) >= self.flood_check_usermark:
				self.flood = True
				self.logger.log("FLOOD DETECTED: %d users in the last %d second(s)!" %
					(current_count-self.last_count, 1), L_STATUS)
			else:
				self.flood = False
				
#		self.logger.log("Floodlog: Flood?: %s, Last: %d, Current: %d" % (
#			self.flood, self.last_count, current_count), L_DEBUG)

		self.last_count = current_count
